/*
**  BootstrapPackage.ox
**
**  Code for replicating results in
**  BOOTSTRAP TESTING OF HYPOTHESES ON CO-INTEGRATING RELATIONS IN VAR MODELS
**
**  General functions for bootstrap testing and Monte Carlo simulation.
**  Written for OX 6.30 and requires the OxMPI/loop class, see www.doornik.com.
**
**  Bootstrap tests are implemented with the function
**      Bootstrap_Test(pr,isMPI,method,iter,Y,EvaluateStatisticFunc,names,GenerateSampleFunc);
**             pr                    : 1 = print output from bootstrap run.
**             isMPI                 : 1 = use MPI.
**             method                : Text string with bootstrap resampling method, "iid" or "wild".
**             iter                  : Integer with number of bootstrap replications.
**             Y                     : Matrix with original data.
**             EvaluateStatisticFunc : Function to evaluation the original/bootstrap statistic using the format EvaluateStatistic(data,isActualData).
**             names                 : Names of the statistics.
**             GenerateSampleFunc    : Function to generate a bootstrap sample using the format GenerateSample(data,shock).
**  see example of usage in Example.ox.
**
**  Monte Carlo simulations are implemented with the function
**      FreeSimulation(pr_all,saveall,mciter,DGP,DGPpar,Simulation_Iteration,names)
**             pr_all                : 1 = iterations  are printed as simulation proceeds
**             saveall               : 1 = Simulation results are saved to file
**             mciter                : Integer with number of Monte Carlo replications
**             DGP                   : Function with monte Carlo DGP using the format MonteCarloDGP(par)
**             DGPpar                : Matrix with all paramter configurations
**             Simulation_Iteration  : Function to perform one Monte carlo replication using the format Simulation_Iteration(data)
**             names                 : array with names of statistics
**  see example of usage in Monte_Carlo_Replication_File.ox.
**
**  Giuseppe Cavaliere, Heino Bohn Nielsen, and Anders Rahbek
**  University of Bologna and University of Copenhagen
**  University of Copenhagen
**  University of Copenhagen
*/


#include <oxstd.h>
#include <oxdraw.h>
#include <oxfloat.h>
#include <oxprob.h>
#import <packages/oxmpi/loop>


class Bootstrap
  {
  /////////////////////////////////////////////////////////////////////////
  //Auxiliary procedures
  static setup_par(const A);
  static PrintProcess(const iRep, const mAdd);

  /////////////////////////////////////////////////////////////////////////
  //Bootstrap variables
  static decl Bootstrap_method;
  static decl Bootstrap_names;
  static decl Bootstrap_data;
  static decl Bootstrap_stat;
  static decl Bootstrap_residual;
  static decl Bootstrap_isMPI;
  static decl Bootstrap_pr;
  static decl Bootstrap_distribution;  
  static decl Bootstrap_para;  
  static decl Bootstrap_results;
  
  /////////////////////////////////////////////////////////////////////////
  //Bootstrap function names (to be overwritten)
  static decl EvaluateStatisticFUNC;
  static decl GenerateSampleFUNC;
  
  /////////////////////////////////////////////////////////////////////////
  //Bootstrap functions
  Bootstrap();
  static Bootstrap_GenerateShocks(const residual);
  static Bootstrap_SimulateDistribution(const iter,creject);
  static Bootstrap_Run(const i);
  static Bootstrap_Test(const pr,const isMPI,const method,const iter,const Y,const EvaluateStatisticFunc,const names,const GenerateSampleFunc);
  static GetParameters();
  static GetBootstrapResults();
  static GetBootstrapDistribution();

  /////////////////////////////////////////////////////////////////////////
  //Simulation variables
  decl Simulation_DGPpar;
  decl Simulation_isMPI;
  static decl Simulation_pr;
  static decl Simulation_SimResults;
  static decl Simulation_CollectedResults;
  static decl Simulation_FullResults;

  /////////////////////////////////////////////////////////////////////////
  //Simulation function names (to be overwritten)
  decl DGPFREEFUNC;
  decl SimulationFREEFUNC;
  
  /////////////////////////////////////////////////////////////////////////
  //Simulation functions
  Simulation_FREERun(const i);
  FreeSimulation(const pr_all,const saveall,const mciter,const DGP,const DGPpar,const Simulation_Iteration,const names);
  static GetCollectedResults();  
  static GetFullResults();  
  static GetSimulation_pr();
  }


//
//class constructor
Bootstrap::Bootstrap()
  {
  Loop::Init();
  Loop::SetSlaveOutput(0);
  print("\n\n***********************************");
  print("\nInitializing Bootstrap Class...");
  }


//
//Build matrix with parameter configurations
Bootstrap::setup_par(const A)
  {
  decl AA,m,M,tal=0;
  decl col = 1;
  
  AA = new array[rows(A)];

  //make ranges for each component
  for(decl i=0;i<rows(AA);++i)
    {
    if(rows(A[i])==1 && columns(A[i])==1)
      {
      AA[i] = A[i];
      }
    else if(rows(A[i])==1 && columns(A[i])>1)
      {
      AA[i] = vec(A[i]);
      }
    else
      {
      AA[i] = vec(range(min(A[i][0:1]),max(A[i][0:1]),A[i][2]));
      }
    col = col*rows(AA[i]);
    }

  //collecting
  M = zeros(rows(A),col);
  m = zeros(rows(A),1);
  //first loop
  for(decl i=0;i<rows(AA[0]);++i)
    {
    m[0] = AA[0][i];

    //second
    if(rows(A)>1)
      {
      for(decl ii=0;ii<rows(AA[1]);++ii)
        {
        m[1] = AA[1][ii];

        //third
        if(rows(A)>2)
          {
          for(decl iii=0;iii<rows(AA[2]);++iii)
            {
            m[2] = AA[2][iii];

            //fourth
            if(rows(A)>3)
              {
              for(decl iiii=0;iiii<rows(AA[3]);++iiii)
                {
                m[3] = AA[3][iiii];

                //fifth
                if(rows(A)>4)
                  {
                  for(decl iiiii=0;iiiii<rows(AA[4]);++iiiii)
                    {
                    m[4] = AA[4][iiiii];

                    //sixth
                    if(rows(A)>5)
                      {
                      for(decl iiiiii=0;iiiiii<rows(AA[5]);++iiiiii)
                        {
                        m[5] = AA[5][iiiiii];

                        //seventh
                        if(rows(A)>6)
                          {
                          for(decl iiiiiii=0;iiiiiii<rows(AA[6]);++iiiiiii)
                            {
                            m[6] = AA[6][iiiiiii];

                            M[][tal] = m;
                            tal = tal+1;
                            }
                          }
                        else
                          {
                          M[][tal] = m;
                          tal = tal+1;
                          }                  
                        }
                      }
                    else
                      {
                      M[][tal] = m;
                      tal = tal+1;
                      }                 
                    }
                  }
                else
                  {
                  M[][tal] = m;
                  tal = tal+1;
                  }      
                }
              }
            else
              {
              M[][tal] = m;
              tal = tal+1;
              }
            }
          } 
        else
          {
          M[][tal] = m;
          tal = tal+1;
          }
        }
      }
    else
      {
      M[][tal] = m;
      tal = tal+1;
      }
    }

  return M;
  }


//
//generate bootstrap shock from residuals
Bootstrap::Bootstrap_GenerateShocks(const residual)
  {
  decl shock,id;

  if(Bootstrap_method=="Wild" || Bootstrap_method=="wild" || Bootstrap_method=="WILD")
    {
    //wild bootstrap (Gaussian distribution)
    shock = residual.*rann(1,columns(residual));
    }
  else if(Bootstrap_method=="iid" || Bootstrap_method=="i.i.d." || Bootstrap_method=="IID")
    {
    //iid bootstrap (random draws with replacement)
    id    = floor(ranu(1,columns(residual))*columns(residual));
    shock = residual[][id];
    }

  //return  
  return shock;
  }

  
//
//one bootstrap iteration
Bootstrap::Bootstrap_Run(const i)
  {
  decl Bootstrap_shock,Bootstrap_sample,Bootstrap_statistic;

  //bootstrap cycle
  Bootstrap_shock     = Bootstrap_GenerateShocks(Bootstrap_residual);
  Bootstrap_sample    = GenerateSampleFUNC(Bootstrap_data,Bootstrap_shock);
  Bootstrap_statistic = EvaluateStatisticFUNC(Bootstrap_sample,FALSE)[0];

  //return
  return Bootstrap_statistic;
  }

  
//
//bootstrap loop
Bootstrap::Bootstrap_SimulateDistribution(const iter,creject)
  {
  decl stat,distribution,reject=0,itno=0;
  
  //place holder
  distribution = constant(.NaN,rows(matrix(Bootstrap_stat)),iter);

  //loop over effective replications
  for(decl i=0;itno<iter;++i)
    {
    //run cycle
    stat = Bootstrap_Run(i);

    //check rejection
    if(stat==<>)
      {
      reject = reject+1;
      }
    else
      {
      distribution[][itno] = stat;
      itno = itno+1;
      }
    }

  //number of rejected
  creject[0] = reject;

  //return
  return distribution;
  }

  
//
//get parameters for bootstrap DGP
Bootstrap::GetParameters()
  {
  return Bootstrap_para;
  }

  
//
//get bootstrap results
Bootstrap::GetBootstrapResults()
  {  
  return Bootstrap_results;
  }

  
//
//get full bootstrap distribution
Bootstrap::GetBootstrapDistribution()
  {  
  return Bootstrap_distribution;
  }

  
//
//main function for bootstrap test
Bootstrap::Bootstrap_Test(const pr,const isMPI,const method,const iter,const Y,const EvaluateStatistic,const names,const GenerateSample)
  {
  decl moment,quant,pval,pval_abs;
  decl Bootstrap_sample;
  decl creject;
  decl tt1,tt2;

  //start timer
  tt1=timer();

  //Using MPI or not?
  #ifdef OX_MPI
    Bootstrap_isMPI = 1;
    //only print at the master
    if(Loop::IsMaster()==TRUE && pr==1)
      {
      Bootstrap_pr = 1;
      }
    else
      {
      Bootstrap_pr = 0;
      }
  #else
    Bootstrap_isMPI = 0;
    Bootstrap_pr    = pr;
  #endif
  
  //if MPI: make slaves silent
  Loop::SetSlaveOutput(0);
  
  //overwrite method
  Bootstrap_method = method;

  //overwrite global functions (to be called from other functions)
  EvaluateStatisticFUNC = EvaluateStatistic;
  GenerateSampleFUNC    = GenerateSample;

  //original data and statistic
  Bootstrap_data   = Y;
  [Bootstrap_stat,Bootstrap_residual,Bootstrap_para]
                   = EvaluateStatisticFUNC(Bootstrap_data,TRUE);

  //headings for statistics
  if(isarray(names)==1)
    {
    Bootstrap_names = names;
    }
  else
    {
    Bootstrap_names = new array[rows(matrix(Bootstrap_stat))];
    for(decl i=0;i<rows(matrix(Bootstrap_stat));++i)
      {
      Bootstrap_names[i] = sprint("Stat(",i,")");
      }
    }
    
  //report
  if(Bootstrap_pr==1)
    {
    print("\n\n***********************************");
    print("\nBootstrap test");
    print("\n\nOriginal data has:","%r",{"p = (rows)","T = (columns)"},rows(Bootstrap_data)|columns(Bootstrap_data));
    print("\nOriginal statistics:");
    print("%r",Bootstrap_names,matrix(Bootstrap_stat));
    
    //check for consistency using residuals
    Bootstrap_sample = GenerateSampleFUNC(Bootstrap_data,Bootstrap_residual);
    print("\nControl of consistency of the generated samples\n   0 =",max(Bootstrap_data-Bootstrap_sample));
    }

  //choose loop construction for MPI or manual
  if(isMPI=="MPI" || isMPI=="mpi")
    {
    //use loop class
    if(Bootstrap_pr==1)    print("\n\nUsing OxMPI and the loop class");
    if(Bootstrap_isMPI==0) print(" (No MPI running...)");

    //settings
    Loop::SetVerbose(0);
    Loop::UseIterSeed(TRUE);        
    Loop::SetSlaveOutput(0);
    //Loop::SetBlockSize(cb);  

    //using MPI loop class
    Bootstrap_distribution = Loop::RunEx(0, Bootstrap_Run , iter, 0, &creject);
    }
  else
    {
    //manual implementation of bootstrap loop
    if(Bootstrap_pr==1) print("\n\nNo MPI");

    //bootstrap loop
    Bootstrap_distribution = Bootstrap_SimulateDistribution(iter, &creject);    
    }
    
  //analyze bootstrap distribution
  moment   = moments(Bootstrap_distribution')';
  quant    = quantiler(Bootstrap_distribution,<.01,.025,.05,.10,.5,.9,.95,.975,.99>);
  pval     = meanr(Bootstrap_distribution.>=Bootstrap_stat);
  pval_abs = meanr(fabs(Bootstrap_distribution).>=fabs(Bootstrap_stat));

  //return global variable
  Bootstrap_results = {Bootstrap_stat',moment',quant',pval',pval_abs'};

  //Master reports
  if(Bootstrap_pr==1)
    {
    print("\n\nBootstrap testing:");
    print("\nBased on ",method," bootstrap with ",iter," replications\n");  
    print("\nRejected = ", creject);
    print("\n","%r",{"actual","Replications","average","std","skewness","kurtosis",".01",".025",".05",".10",".5",".9",".95",".975",".99","p-val","p-val(abs)"},
          "%c",Bootstrap_names,
          (Bootstrap_stat~moment~quant~pval~pval_abs)');

    //time
    tt2=timer();
    print("\nTotal Bootstrap time: ",timespan(tt1,tt2),".\n");        
    }
  }


//
//one Monte Carlo iteration
Bootstrap::Simulation_FREERun(const i)
  {
  decl Y,stat;
    
  //generate data
  Y = DGPFREEFUNC(Simulation_DGPpar);

  //call procedure to calculate MC statistics
  stat = SimulationFREEFUNC(Y);

  //return
  return stat;
  }

  
//
//Free Monte Carlo Simulation
Bootstrap::FreeSimulation(const pr_all,const saveall,const mciter,const DGP,const DGPpar,const Simulation_Iteration,const names)
  {
  decl moment,quant,rest,MCnames;  
  decl t1,t2;
  decl tt1,tt2;

  //start overall timer
  tt1=timer();
  
  //make slaves silent
  Loop::SetSlaveOutput(0);

  //Using MPI or not?
  #ifdef OX_MPI
    Simulation_isMPI = 1;
    //this is the Master
    if(Loop::IsMaster()==TRUE)
      {
      Simulation_pr = 1;
      }
    else
      {
      Simulation_pr = 0;
      }
  #else
    Simulation_isMPI = 0;
    Simulation_pr    = 1;
  #endif

  //Master reports
  if(Simulation_pr==1)
    {
    print("\n\n***********************************");
    print("\nMonte Carlo simulation");
    print("\nUsing OxMPI and the loop class");
    if(Simulation_isMPI==0) print(" (No MPI running....)"); 
    print("\n",mciter," replications");
    print("\n\nParameter configurations:",DGPpar);

    //collect all the results
    Simulation_CollectedResults = new array[columns(DGPpar)];   
    Simulation_FullResults      = new array[columns(DGPpar)];   
    }

  //overwrite global functions
  DGPFREEFUNC              = DGP;
  SimulationFREEFUNC       = Simulation_Iteration;

  //loop over parameter configurations
  for(decl u=0;u<columns(DGPpar);++u)
    {
    decl creject;

    //time each configuration
    t1=timer();
    
    //common random numbers over configurations
    Loop::SetMasterSeed(-1);

    //some global variables
    Simulation_DGPpar = DGPpar[][u];
    
    ////////////////////////////////
    //always use Loop class

    //settings
    Loop::SetVerbose(0);
    Loop::UseIterSeed(TRUE);      
    Loop::SetSlaveOutput(0);
    //Loop::SetBlockSize(1);  

    if(pr_all==0)
      {
      //using MPI loop class
      Simulation_SimResults = Loop::RunEx(0, Simulation_FREERun , mciter, 0, &creject);
      }
    else
      {
      //using MPI loop class, print each batch
      print("\n\nResults are printed as simulation proceeds:");
      Simulation_SimResults = Loop::RunEx(0, Simulation_FREERun , mciter, PrintProcess, &creject);
      }
      
    //Master reports
    if(Simulation_pr==1)
      {
      moment   = moments(Simulation_SimResults');
      quant    = quantiler(Simulation_SimResults,<.01,.025,.05,.10,.5,.9,.95,.975,.99>)';

      //headings for statistics
      if(isarray(names)==1)
        {
        MCnames = names;
        }
      else
        {
        MCnames = new array[rows(matrix(moment'))];
        for(decl i=0;i<rows(matrix(moment'));++i)
          {
          MCnames[i] = sprint("Stat(",i,")");
          }
        }
      
      print("\n\n* * * * * * * * * * * *");
      print("\nConfiguration",DGPpar[][u]);
      print("\nRejected = ", creject);        
      print("%r",{"replications","average","std","skewness","kurtosis",".01",".025",".05",".10",".5",".9",".95",".975",".99"},
            "%c",MCnames,
            (moment|quant));

      //collect all the results
      Simulation_CollectedResults[u] = (moment|quant);
      Simulation_FullResults[u]      = Simulation_SimResults';
          
      //save full set of results      
      if(saveall==1)
        {
        decl d0,d1,filename;

        //create filename     
        d0 = strtrim(date());if(columns(d0)==9) d0 = "0"+d0;
        d1 = strtrim(time());if(columns(d1)==7) d1 = "0"+d1;
        filename = sprint("SIM_",d0[6:]+"_"+d0[3:4]+"_"+d0[0:1]+"_at_"+d1[0:1]+"_"+d1[3:4]);      
        for(decl j=0;j<rows(Simulation_DGPpar);++j)
          {
          filename = filename+sprint("_P[",j,"]_",Simulation_DGPpar[j][0]);
          }
        print("\nSimulation results are saved in ",filename,".xls","  \n");

        //save
        savemat(sprint(filename,".xls"),Simulation_SimResults',MCnames); 
        }
      }

    if(Simulation_pr==1)
      {   
      //time
      t2=timer();
      print("\nTime for this configuration: ",timespan(t1,t2),".\n");
      }
    }
    
  if(Simulation_pr==1)
    {
    //time
    tt2=timer();
    print("\nTotal Monte Carlo time: ",timespan(tt1,tt2),".\n");
    }
  }


//
//get array of collected results
Bootstrap::GetCollectedResults()
  {  
  return Simulation_CollectedResults;
  }

//
//get array of full results
Bootstrap::GetFullResults()
  {  
  return Simulation_FullResults;
  }

//
//print results during run
Bootstrap::PrintProcess(const iRep, const mAdd)
  {
  print(mAdd');

  return mAdd;
  }
  