/*
**  Example.ox
**
**  Code for replicating results in
**  BOOTSTRAP TESTING OF HYPOTHESES ON CO-INTEGRATING RELATIONS IN VAR MODELS
**
**  Example implementing a bootstrap test on the co-integrating relation in a VAR model.
**  Written for OX 6.30 and requires the OxMPI/loop class, see www.doornik.com.
**
**  1) The reduced rank regression of the co-integrated VAR is implemented with the functions i1_datamat() and i1_ana(), see the file i1_estimation.ox.
**  2) The asymptotic LR test and the Bartlett correction is implemented with the function i1_lr_bartlett(), see the file i1_estimation.ox.
**  3) The bootstrap test is implemented using the Bootstrap package, see the file BootstrapPackage.ox.
**     The test specifies EvaluateStatistic(data,isActualData) and GenerateSample(data,shock).
**
**  Giuseppe Cavaliere, Heino Bohn Nielsen, and Anders Rahbek
**  University of Bologna and University of Copenhagen
**  University of Copenhagen
**  University of Copenhagen
*/


#include <i1_estimation.ox>
#include <BootstrapPackage.ox>


//
//global variables
//
decl H,i_intercept;  


//
//evaluate statistic with BOOTSTRAP DGP ESTIMATED UNDER THE NULL
//
//  if isActualData==TRUE, also return residuals and estimates
//
EvaluateStatistic(const data,const isActualData)
  {
  decl stat,residual,residual_u,parameters;

  //estimation
  xx       = data;
  x1       = 0;
  if(i_intercept==0)
    {
    dx1      = 0;
    }
  else
    {
    dx1      = ones(1,columns(data));
    }

  //Supply information for the estimation   
  startper =    {1,1};
  freq     =        1;
  k        =        1;
  start    =      k+1; 
  end      =   columns(xx);
  
  //construct data matrices
  i1_datamat(0);

  if(isActualData==TRUE)
    {
    //also estimate under H
    i1_ana(0,4,0,3,1,1,0,0,0,0);
    //unrestricted residuals
    //residual_u = zeros(p,k)~conc(i1_residual,ones(1,n));

    i1_ana(0,4,0,3,1,1,0,0,1,H);
    test(0,i1_egenv_r,i1_egenv_u);

    //statistic
    stat = btest[0][0];

    //centered restricted residuals
    residual = zeros(p,k)~conc(i1_residual,ones(1,n));

    //parameters under null
    parameters = {i1_pi,i1_exo};

    //return
    return {stat , residual , parameters };
    }
  else
    {
    //only test
    i1_ana(0,4,0,3,1,1,0,0,0,0);
    i1_ana(0,4,0,3,1,1,0,0,1,H);
    test(0,i1_egenv_r,i1_egenv_u);

    //statistic
    stat = btest[0][0];
    
    //return
    return {stat , 0 , 0 };
    }       
  }


//
//generate a new bootstrap sample
//
//  may use function Bootstrap::GetParameters() to get saved parameters
//
GenerateSample(const data,const shock)
  {
  decl sample,pi,mu,dx;

  //parameters 
  pi     = Bootstrap::GetParameters()[0];
  if(i_intercept==1)
    {
    mu     = Bootstrap::GetParameters()[1];
    }

  //generate
  sample = data;
  for(decl i=1;i<columns(data);++i)
    {
    if(i_intercept==0)
      {
      dx          = pi*sample[][i-1] + shock[][i];
      }
    else
      {
      dx          = mu + pi*sample[][i-1] + shock[][i];
      }
    
    sample[][i] = dx + sample[][i-1];
    }

  //return  
  return sample;  
  }


///////////////////////////////////////////////////////////////////////  
//  
//main program block
//
main()
  {
  format(2000);
  decl Y;
  decl i_rank;

  
  //////////////////////////
  i_intercept = 1;     // =1 to allow intercept
  //////////////////////////

  
  //matrix with data
   Y = <      0.0000,      0.2249,     -0.8052,     -0.1792,      0.5907,      1.1089,      0.0541,     -1.4304,     -0.1361,     -0.6574,      0.4083,      1.4324,      1.1663,      1.6485,      0.6075,      0.9609,      0.9415,      1.6652,      0.6558,      0.4052,      0.1294,     -0.5227,     -0.7997,     -0.0227,      0.3480,     -0.6008,     -0.6372,     -1.6886,      0.2274,      1.4519,      2.8757,      1.5537,     -0.7144,     -0.6153,     -0.3306,      0.3233,      0.4928,      1.4279,     -0.1868,     -0.6489,     -0.0558,     -0.8660,     -1.0627,     -0.9826,     -1.3807,     -0.9501,      0.0399,     -1.1191,     -0.6454,      0.0961,     -0.5891,     -2.5140,     -1.4006,     -1.0012,     -1.4706,     -0.7479,     -0.0656,      0.8151,      1.8503,      1.9510,      2.3294,      0.9661,      1.5975,      0.5134,      0.2047,      1.7010,     -0.6797,      1.1089,      1.1439,      1.6790,      0.9713,     -1.0454,     -1.4970,     -0.4549,      0.0185,     -1.2132,     -1.4488,      0.2422,     -0.7346,     -1.3059,     -0.3667,      0.9700,      0.9440,      0.2990,     -0.6451,     -0.2210,     -0.3685,     -0.8782,     -2.6586,     -1.9239,     -0.4623,      1.1982,      1.1341,     -0.6794,     -1.1286,     -1.2925,     -0.9373,     -1.0164,     -1.5275,     -0.5598;
              0.0000,      1.7400,      1.0659,      0.9245,      1.0150,      2.1324,      0.5057,     -1.2978,     -2.0058,     -0.5384,      0.9411,      2.5195,      3.8009,      1.8465,      2.2531,      2.0862,      2.2016,      0.7416,     -0.1743,     -1.0090,     -1.7444,     -2.3390,     -4.5137,     -4.3786,     -5.9035,     -6.4551,     -6.8043,     -6.9345,     -6.9726,     -7.4727,     -6.5247,     -6.2027,     -6.8882,     -7.5289,     -7.3637,     -8.1593,     -7.8047,     -7.5520,     -7.6080,     -8.0469,     -8.3192,    -10.1234,     -7.3568,     -5.9631,     -8.6708,     -8.9782,     -8.3170,     -6.8112,     -5.8587,     -6.0222,     -7.4128,     -5.1015,     -4.5108,     -6.3778,     -8.7515,     -8.9411,     -6.3185,     -7.8618,     -9.1540,     -9.0264,     -9.2525,     -9.3252,     -8.9548,     -9.0203,    -10.0708,    -11.3139,    -10.0637,    -10.3832,     -9.6330,     -8.7864,     -7.9007,     -6.9317,     -8.0096,     -7.6418,     -6.8617,     -7.9252,     -7.1835,     -8.7969,     -9.0090,     -9.3354,     -9.2784,    -10.8755,    -11.1575,    -13.2822,    -14.3104,    -12.8692,    -10.8554,    -10.9266,     -9.0004,     -8.1020,     -6.5459,     -5.4509,     -5.7690,     -4.0050,     -4.7775,     -4.7171,     -5.8438,     -5.3963,     -6.0717,     -6.3867;
              0.0000,     -0.2043,     -0.5478,     -0.7312,     -1.5644,     -1.2495,     -0.6300,      1.3715,      1.9649,      1.9447,      1.6758,      1.2825,      0.9208,      0.8188,      0.9571,      1.4355,      1.4735,      1.8459,      2.3154,      2.2053,      1.9476,      3.0825,      4.2470,      4.4823,      4.3635,      3.8579,      3.3191,      3.5210,      1.9872,      1.9737,      2.5620,      2.6562,      1.0747,      0.6335,      0.6805,      0.4499,      1.6317,     -0.4664,     -1.3437,     -0.4949,     -0.5831,     -0.6891,      0.0436,      0.0013,     -0.3405,      1.0519,     -0.3203,      0.3893,      0.6949,      0.6681,      0.9081,      2.3510,      1.6054,      0.0384,      1.9020,      0.9679,      2.2179,      1.8338,      2.6143,      2.2857,      2.8690,      2.7083,      3.7902,      3.2618,      2.5581,      1.3438,      1.5010,      0.9894,      0.6473,      0.2414,     -1.8819,     -1.3860,     -3.9105,     -3.0299,     -2.3795,     -3.2878,     -3.0725,     -3.9768,     -4.2699,     -5.8445,     -7.2072,     -5.9843,     -5.4549,     -5.0477,     -7.9694,     -7.4901,     -8.9432,     -8.6332,     -6.8695,     -5.2455,     -5.2099,     -5.7970,     -5.4823,     -5.9234,     -6.6361,     -7.3095,     -7.6357,     -8.1307,     -6.4884,     -6.0405>;

              
  //Hypothesis beta=H
  H          = <1;
                0;
                0>;

  
  //global variables for estimation
  xx       = Y;
  x1       = 0;
  if(i_intercept==0)
    {
    dx1      = 0;
    }
  else
    {
    dx1      = ones(1,columns(Y));
    }

  //Supply information for the estimation   
  startper =    {1,1};
  freq     =        1;
  k        =        1;
  start    =        k+1; 
  end      =   columns(xx);
  
  //generate data matrices for estimation
  i1_datamat(1);

  //unrestricted estimation
  i1_ana(1,4,0,2,1,1,0,0,0,0);

  //restricted estimation
  i1_ana(1,4,0,2,1,1,0,0,1,H);

  //Bartlett corrected test
  if(i_intercept==0)
    {
    i1_lr_bartlett(1,1,0,H);
    }
  else
    {
    i1_lr_bartlett(1,2,2,H);
    }

  //Proposed bootstrap
  decl myseed = ranseed(0);
  Bootstrap::Bootstrap_Test(1,"noMPI","iid",999,Y,EvaluateStatistic,0,GenerateSample);
  }
                              