/*
**  I1_Estimation.ox
**
**  Code for replicating results in
**  BOOTSTRAP TESTING OF HYPOTHESES ON CO-INTEGRATING RELATIONS IN VAR MODELS
**
**  This code contains code for reduced rank regression, likelihood ratio test, and Bartlett correction.
**
**  The main functions are:
**
**  1) i1_datamat(pr)
**     Produces data matrices for estimation (pr=1 reports settings)
**     Needs the following global variables
**          xx        : matrix with data in rows
**          x1        : matrix with restricted deterministics
**          dx1       : matrix with unrestricted deterministics
**          startper  : array with integer for year and month/quarter
**          freq      : interger with number of perods per year
**          k         : integer with number of lags
**          start     : integer with first effective observation in xx
**          end       : integer with last effective observation in xx
**  2) i1_ana(pr,algo,ranktest,det,param,rr,ralpha,ra,rbeta,rb)
**     Reduced rank regression of the co-integrated VAR 
**          pr        : (0/1)     print outpt or not
**          algo      : (1/4)     algoritm for eigenvalue problem 1=eigen,4=singular values
**          ranktest  : (0/1)     calculate rank test or not
**          det       : (0/1/2/3) deterministic specification 0=none,1=restr.const,2=unrstr.const,3=restr.trend
**          param     : (0/1)     estimate paramters in H(rr)
**          rr        :           co-integration rank
**          ralpha    : (0/1)     linear restrictions on alpha
**          ra        :           matrix with restrictions on alpha, alkpha=H*phi
**          rbeta     : (0/1)     linear restrictions on all columns in beta
**          rb        :           matrix with restrictions, beta_star=H*phi
**  3) test(pr,er,eu)
**     Likelihood ratio statistic computed by comparing restricted and unrestridted eigenvalues
**          pr        : (0/1)     print outpt or not
**          er        :           matrix with restricted eigenvalues
**          eu        :           matrix with unrestricted eigenvalues
**  4) i1_lr_bartlett(pr,rr,det,restr)
**     Bartlett corrected test of hypothesis on beta
**          pr        : (0/1)     print outpt or not
**          rr        :           co-integration rank
**          det       : (0/1/2/3) deterministic specification 0=none,1=restr.const,2=unrstr.const,3=restr.trend
**          rb        :           matrix with restrictions, beta_star=H*phi
**
**  Giuseppe Cavaliere, Heino Bohn Nielsen, and Anders Rahbek
**  University of Bologna and University of Copenhagen
**  University of Copenhagen
**  University of Copenhagen
*/


#include <oxstd.h>
#include <oxprob.h>
#include <../lib/coigamma.ox>
#include <oxdraw.h>
#include <oxfloat.h>
#import  <maximize>
#import  <maxsqp>


//
//i1_datamat(pr)
//constructs matrices for I(1) estimation
//   pr : (0/1) report settings or not
//
decl xx,x1,xa,dx1;
decl k,startper,start,end,freq,txx;
decl kexo = -1;
decl n,in,p,p_r,p_u;
decl zz0,zz2,zz2_star,zz3,zz7;
decl talaxis,axis;
decl p_zz3;
decl ksparam,n_param,df,btest,bpval;
i1_datamat(const pr)
  {  
  format(2000);

  decl i,j,var;
  decl varre,temp,temp1,tal,tal1;
  decl p1,p2;
  decl t;
  decl indi;
  decl k_exo;

  indi = 0;
  tal  = 0;
  tal1 = 0;

  //Define matrices with levels of the variables
  var = {xx[][start-1:end-1]};  
  for (i = 1 ; i <= k ; ++i)
    {    
    var = var~xx[][start-1-i:end-1-i];
    }

  //Setup zz0 and zz2
  zz0      = var[0]-var[1];
  zz2      = var[1];
  zz2_star = zz2;

  //Setup the stochastic part of zz3.
  //In case k=1 we choose something... and delete the row afterwards
  if(k>1)
    {
    zz3 = var[1]-var[2];
    if(k>2)
      {
      for (i = 2 ; i <= k-1 ; ++i)
        {
        zz3 = zz3|(var[i]-var[i+1]);
        }
      }
    }
  else if(k==1)
    {
    zz3  = sqrt(range(1,columns(zz0)));
    indi = 1;
    }

  //use kexo lags.
  if(kexo==-1)
    {
    k_exo=k;
    }
  else
    {
    k_exo = kexo;
    }
  
  //Setup restricted deterministic variables  
  if (x1==0)
    {
    zz7 = 0;      
    //print("No restricted deterministic variables");
    }
  else
    {
    //Matrices with levels of the restricted exogenous
    varre = {x1[][start-1:end-1]};  
    for (i = 1 ; i <= k_exo ; ++i)
      {
      varre = varre~x1[][start-1-i:end-1-i];
      }

    //Augment zz3 with linear independent contributions
    for (j = 0 ; j<=k_exo-1 ; ++j)
      {
      temp = varre[j]-varre[j+1];
      for (i = 0 ; i<=rows(temp)-1 ; ++i)
        {
        temp1 = temp[i][];
        //Test if row is zero
        if(rank(temp1)==1)
          {
          //Test if row is independent of zz3
          if(rank(zz3|temp1)==(rows(zz3)+1))
            {
            zz3  = zz3|temp1;
            tal1 = tal1+1;
            }
          }
        }  
      }
      
    //Setup the levels if they are linearly independent of zz3
    temp = varre[1];
    for (i = 0 ; i<=rows(temp)-1 ; ++i)
      {
      temp1 = temp[i][];
      //Test if row is zero
      if(rank(temp1)==1)
        {
        //Test if row is independent of zz3
        if(rank(zz3|temp1)==(rows(zz3)+1))
          {
          zz2_star = zz2_star|temp1;
          }
        }
      }
      
    //Calculate deterministic matrix  
    zz7 = zz2_star[rows(xx):][];      
    }

  //Setup unrestricted deterministic variables
  if (dx1==0)
    {
    //print("No unrestricted deterministic variables");
    }
  else
    {
    temp = dx1[][start-1:end-1];
    for (i = 0 ; i<=rows(temp)-1 ; ++i)
      {
      temp1 = temp[i][];
      //Test if row is zero
      if(rank(temp1)==1)
        {
        //Test if row is independent of zz3
        if(rank(zz3|temp1)==(rows(zz3)+1))
          {
          zz3 = zz3|temp1;
          tal = tal+1;
          }
        }
      }     
    }   

  //Remove a columns of random numbers if k=1
  if (indi==1)
    {
    if (rows(zz3)==1)
      {
      zz3 = 0;
      }
    else
      {
      zz3 = zz3[1:][];
      }
    }
  
  //Calculate constants  
  n     = columns(zz0);
  in    = 1.0/n;
  p     = rows(zz0);
  p_r   = rows(zz7);
  p_u   = tal+tal1;
  p_zz3 = (k-1)*p+p_u;
  
  if(pr==1) print("\n \n");
  if(pr==1) print("************************************************************************** \n \n"); 
  if(pr==1) print("I(1) analysis \n\n");
  if(pr==1) print("Number of lags:                       ",k       ,"\n");
  if(pr==1) print("Number of effective observations:     ",n       ,"\n");  
  if(pr==1) print("Number of endogenous variables:       ",p       ,"\n");  
  if(pr==1) print("Deterministic: Restricted to levels:  ",p_r     ,"\n");  
  if(pr==1) print("               Unr. differences:      ",(k-1)*p ,"\n");
  if(pr==1) print("               Unrestricted:          ",p_u     ,"\n\n");

  //Constructing an explicit time axis

  if(startper[1]==0)
    {
    axis       = new array[end-start+1][1];
    for (i=0 ; i<=end-start ; ++i)
      {
      axis[i] = sprint("%d", i);
      }
    }
  else
    {
    t = range(1,end-start+1);

    talaxis       = new matrix[end-start+1][1];
    talaxis[0][0] = startper[0]+fmod((startper[1]-1)/freq,1);

    axis          = new array[columns(t)];
    p1 = floor(talaxis[0]+0.01);
    p2 = 1+round(freq*(talaxis[0]-floor(talaxis[0]+0.01)));
    axis[0]    = sprint("%d", p1 ,":","%d", p2 );
  
    for (i=1 ; i<=columns(t)-1 ; ++i)
      {
      talaxis[i] = talaxis[i-1]+1.0/freq;
      p1 = floor(talaxis[i]+0.01);
      p2 = 1+round(freq*(talaxis[i]-floor(talaxis[i]+0.01)));
      axis[i]    = sprint("%d", p1 ,":","%d", p2 );
      }
    if(pr==1) print("Effective sample: ", axis[0]," - ",axis[columns(t)-1],"\n");
    }
  }

  
//
//least squares concentration
//
conc(const ind, const bet)
  {
  decl pi,ret;

  if(bet==0)
    {
    ret = ind;
    }
  else if(bet==<>)
    {
    ret = ind;
    }
  else
    {
    olsr(ind , bet , &pi);
    ret = ind - pi*bet;
    }
    
  return ret;
  }


//
//algorithm for solving the eigenvalueproblem
//
eigsolve(const algo, const rr0, const rr2, eigv, alp, bet)
  {
  decl y0,y1,eigval,eigvec;

  if (algo==1)
    {
    //Algorithm 1
    //Generalized eigenvalueproblem
    decl ss00,ss02,ss22;
  
    ss00 = in*rr0*rr0';
    ss02 = in*rr0*rr2';
    ss22 = in*rr2*rr2';
  
    eigensymgen(ss02'*invert(ss00)*ss02, ss22, &eigval, &eigvec);

    eigv[0] = eigval;
    bet[0]  = eigvec;
    alp[0]  = ss02*bet[0];
    }

  if (algo==4)
    {
    //Algorithm 4
    //Based on SVD.
    y0 = rr0';
    y1 = rr2';
    decl u0,w0,v0,u1,w1,v1,z,uz,wz,vz;

    decsvd(y0,&u0,&w0,&v0);
    decsvd(y1,&u1,&w1,&v1);

    z = u1'u0;
    decsvd(z,&uz,&wz,&vz);  

    eigv[0] = wz.^(2);
    bet[0]  = n^(.5)*v1*diag(w1.^(-1))*uz;
    alp[0]  = n^(-.5)*y0'u1*uz;
    }
  }


//
//i1_ana(pr,algo,ranktest,det,param,rr,ralpha,ra,rbeta,rb)
//I(1) reduced rank estimation
//   pr       : (0/1)     print outpt or not
//   algo     : (1/4)     algoritm for eigenvalue problem 1=eigen,4=singular values
//   ranktest : (0/1)     calculate rank test or not
//   det      : (0/1/2/3) deterministic specification 0=none,1=restr.const,2=unrstr.const,3=restr.trend
//   param    : (0/1)     estimate paramters in H(rr)
//   rr       :           cointegration rank
//   ralpha   : (0/1)     linear restrictions on alpha
//   ra       :           matrix with restrictions on alpha, alkpha=H*phi
//   rbeta    : (0/1)     linear restrictions on all columns in beta
//   rb       :           matrix with restrictions, beta_star=H*phi
//
decl r;
decl fri;        
decl i1_tracetest,pval,i1_lmaxtest,pval2;
decl i1_beta_star,i1_beta,i1_alpha,i1_pi,i1_gamma_mat,i1_t_gamma_mat,i1_exo;
decl i1_t_alpha,i1_t_exo;
decl i1_residual,i1_omega,i1_loglik,i1_egenv_u,i1_egenv_r;
decl r_hat,i1_det;
decl i1_tracetest_quan;
decl eigval;
i1_ana(const pr,const algo,const ranktest,const det,const param,const rr,const ralpha,const ra,const rbeta,const rb)
  {
  decl i;
  decl zz2_mat,r0,r2_star,s00,s02_star,s22_star,antalgamma;
  decl xxksparam,varksparam;
  decl a_fri,b_fri;
  decl ra_bar,ra_orto,r0b;
  decl gammapar;
  decl alpha2,beta_star2,residuals;
  decl alpha,beta_star;

  if(pr==1) print("\n \n");
  if(pr==1) print("************************************************************************** \n \n"); 
  if(pr==1) print("I(1) analysis \n \n");

  //Impose potential restrictions on Beta_star
  fri = 0;
  if (ralpha==1)
    {
    if(pr==1) print("Restricted estimation: Alpha=A*Phi  \n");
    if(pr==1) print("A",ra);

    zz2_mat = zz2_star;
    a_fri   = rows(ra)-columns(ra);
    fri     = fri + a_fri;
    }
  if (rbeta==1)
    {
    if(pr==1) print("Restricted estimation: Beta_star=B*Phi \n");
    if(pr==1) print("B",rb);

    zz2_mat = rb'zz2_star;
    b_fri   = rows(rb)-columns(rb);
    fri     = fri + b_fri;
    }
  if (ralpha==0 && rbeta==0)
    {
    if(pr==1) print("Unrestricted estimation");
    zz2_mat = zz2_star;
    }

  //Find concentrated RRR model
  if (p_zz3>0)
    {
    r0      = conc(zz0,zz3);
    r2_star = conc(zz2_mat,zz3);
    }
  else
    {
    r0      = zz0;
    r2_star = zz2_mat;
    }

  //Impose the potential retriction of Alpha
  if (ralpha==1)
    {
    ra_bar  = ra*invert(ra'ra);
    ra_orto = nullspace(ra);

    //Transform the system with ra_bar
    r0b      = r0;
    r0       = conc(ra_bar'r0b, ra_orto'r0b);
    r2_star  = conc(r2_star   , ra_orto'r0b);
    }
  else
    {
    ra_bar = unit(p);
    }

  //Solve the eigenvalueproblem and print eigenvalues
  eigsolve(algo,r0,r2_star,&eigval,&alpha2,&beta_star2);  
  if(pr==1) print("%r",{"Eigenvalues:           "},eigval);
  if(pr==1) print("Based on algorithm ",algo,"\n\n");

  //Recover LS parameters (full rank case)  
  if (rbeta==1 && ralpha==1)
    {
    alpha      = ra_bar*alpha2;
    beta_star  = rb*beta_star2;
    i1_egenv_r = eigval;
    }
  if (rbeta==1 && ralpha==0)
    {
    alpha      = alpha2;
    beta_star  = rb*beta_star2;
    i1_egenv_r = eigval;
    }
  if (rbeta==0 && ralpha==1)
    {
    alpha      = ra_bar*alpha2;
    beta_star  = beta_star2;
    i1_egenv_r = eigval;
    }
  if (rbeta==0 && ralpha==0)
    {
    alpha      = alpha2;
    beta_star  = beta_star2;
    i1_egenv_u = eigval;
    }

  decl gammapar2,quan1,quan2;   
  //Calculate the Trace test and the assymptotic critical values
  if (ranktest==1)
    {
    i1_tracetest = -n*reversec(cumulate(log(1-reversec(eigval[][0:p-1]'))))';
    gammapar     = GammaParTraceAsymp(det, range(p,1)', 0);
    pval         = 1 - probgamma(i1_tracetest', gammapar[][0], gammapar[][1]);
    quan1        = quangamma(0.95, gammapar[][0], gammapar[][1])';
    i1_tracetest_quan = quan1;

    i1_lmaxtest  = -n*log(1-eigval[][0:p-1]);
    gammapar2    = GammaParMaxEvAsymp(det, range(p,1)', 0);

    pval2        = 1 - probgamma(i1_lmaxtest', gammapar2[][0], gammapar2[][1]);
    quan2        = quangamma(0.95, gammapar2[][0], gammapar2[][1])';
        
    if(pr==1) print("************************************************************************** \n \n");
    if(pr==1) print("Rank determination in I(1) model \n\n");
    if(pr==1) print("%r",{"H0: r=",
                          "Eigenvalues",
                          "Trace test statistics",
                          "95 pct. critical values",
                          "Asymptotic p values",
                          "LambdaMax statistics",
                          "95 pct. critical values",
                          "Asymptotic p values"},
              (cumulate(ones(p,1))'-1)|
              (eigval[][0:p-1])|
              (i1_tracetest)|
              quan1|
              (pval')|
              i1_lmaxtest|
              quan2|
              (pval2')
              );
    if (det==0 && pr==1) print("\nNote: No deterministic varaibles\n");
    if (det==1 && pr==1) print("\nNote: Restricted constant\n");
    if (det==2 && pr==1) print("\nNote: Unrestricted constant\n");
    if (det==3 && pr==1) print("\nNote: Restricted linear trend term\n");

    //adhoc: save deterministic setting
    i1_det = det;
    
    //adhoc: estimate r
    for(i=0;i<=p-1;++i)
      {
      if(pval[i][0]>0.05)
        {
        r_hat = i;
        break;
        }
      else r_hat=p;  
      }
    }

  //Calculate all parameters given r
  if (param==1)
    {
    //Note:everything here is calculated from r0,r2 and the folded estimates alpha2,beta_star2

    r = rr;

    if(pr==1) print("************************************************************************** \n \n");
    if(pr==1) print("Estimated parameters \n");
    if(pr==1) print("r=",r,"\n");

    if(r>0)
      {
      i1_beta_star = beta_star[][:rr-1]; 
      i1_beta      = beta_star[0:p-1][:rr-1];
      i1_alpha     = alpha[][:rr-1];
      i1_pi        = i1_alpha*i1_beta_star';

      if(pr==1) print("Beta_star (transposed)",i1_beta_star');
      if(pr==1) print("Alpha"                 ,i1_alpha);

      //calculate variance of alpha
      residuals  = r0-alpha2[][0:r-1]*beta_star2[][0:r-1]'r2_star;

      olsr(r0,beta_star2[][0:r-1]'r2_star, &ksparam, &xxksparam);   
      i1_t_alpha = i1_alpha./((ra_bar*shape(diagonal(xxksparam**(in*residuals*residuals'))',rows(r0),r)).^.5);
      if(pr==1) print("t-value"               ,i1_t_alpha);     
      if(pr==1) print("Pi"                    ,i1_pi);    
      }
    else
      {
      i1_beta_star = zeros(rows(zz2_star),1); 
      i1_beta      = zeros(p,1);
      i1_alpha     = zeros(p,1);
      i1_pi        = i1_alpha*i1_beta_star';      

      if(pr==1) print("Beta_star (transposed)",i1_beta_star');
      if(pr==1) print("Alpha"                 ,i1_alpha);
      if(pr==1) print("Pi"                    ,i1_pi);    
      }

    //Short run parameters
    if (columns(zz3)>0)
      {
      //OLS estimation
      olsr(conc(zz0,i1_beta_star'zz2_star), conc(zz3,i1_beta_star'zz2_star), &ksparam, &xxksparam);

      //finding residuals
      i1_residual  = zz0-i1_alpha*i1_beta_star'zz2_star-ksparam*zz3;
      
      //variance estimates
      varksparam = shape(diagonal(xxksparam**(in*i1_residual*i1_residual'))',rows(ksparam),columns(ksparam));

      antalgamma = k-1;
      if (antalgamma>0)
        {
        i1_gamma_mat   = new array[antalgamma];
        i1_t_gamma_mat = new array[antalgamma];
        for (i = 1 ; i <= antalgamma ; ++i)
          {
          i1_gamma_mat[i-1]   = ksparam[][(i-1)*p:i*p-1];
          i1_t_gamma_mat[i-1] = (i1_gamma_mat[i-1])./(varksparam[][(i-1)*p:i*p-1]).^.5;
          if(pr==1) print("Gamma(",i,")"        ,i1_gamma_mat[i-1]);
          if(pr==1) print("t-values"            ,i1_t_gamma_mat[i-1]);        
          }
        }
      if (columns(ksparam)>antalgamma*p)
        {
        i1_exo   = ksparam[0:p-1][antalgamma*p:columns(ksparam)-1];
        i1_t_exo = i1_exo./(varksparam[0:p-1][antalgamma*p:columns(ksparam)-1]).^.5;
        //if(pr==1) print("Exogenous"        ,i1_exo);
        //if(pr==1) print("t-values"         ,i1_t_exo);
        decl help = 0;
        while(help<columns(i1_exo))
          {
          if(pr==1) print("Exogenous",i1_exo[][help:min(help+4,columns(i1_exo)-1)]);
          if(pr==1) print("t-values" ,i1_t_exo[][help:min(help+4,columns(i1_exo)-1)]);
          help = help + 5;
          }        
        }        
      }
    else
      {
      //finding residuals
      i1_residual  = zz0-i1_alpha*i1_beta_star'zz2_star;
      }

    //loglikelihood value
    i1_omega  = in*i1_residual*i1_residual';
    if(pr==1) print("Covariance matrix","%lwr",i1_omega);

    i1_loglik = -0.5*n*log(determinant(i1_omega));
    if(pr==1) print("Maximum value of the log-likelihood function \n");
    if(pr==1) print("loglik=",i1_loglik,"    (",i1_loglik-0.5*n*p*log(M_2PI)-0.5*p*n,")\n \n");
    }
  }

  
//
//LR tests based on two sets of eigenvalues
//
test(const pr,const er,const eu)
  {
  decl eer,eeu,ra;

  //Calculate eigenvalues strippet for zeroes
  eer = deleteifc(er,isdotfeq(er, zeros(rows(er),columns(er))));
  eeu = deleteifc(eu,isdotfeq(eu, zeros(rows(eu),columns(eu))));

  eer = er;
  eeu = eu;

  if(pr==1) print("\n \n");
  if(pr==1) print("************************************************************************** \n \n"); 
  if(pr==1) print("Eigenvalues in two estimations");
  if(pr==1) print("%r",{"Unrestricted:  "},eeu);
  if(pr==1) print("%r",{"Restricted:    "},eer);

  btest = n*cumulate(  (log((1-eer)./(1-eeu[][0:columns(eer)-1])))'  );
  ra    = range(1,columns(eer),1)';
  bpval = constant(1,columns(eer),1)-probchi(btest,fri*ra);

  if(pr==1) print("\nTest statistics for different values of r");  
  if(pr==1) print("%c",{"r or s","LR-test","p-value","df"},ra~btest~bpval~fri*ra);  
  }

  
//
//Bartlett correction of test on beta of the form Beta = H*Phi
//
decl correction;
decl fpval;
i1_lr_bartlett(const pr,const rr,const det,const restr)
  {
  //Bartlett correction for test of hypothesis on Beta  
  decl i;
  decl rang,i1_alpha_orto;
  decl ny,s,nD,nd;
  decl m;
  decl bigp,bigq;
  decl rakke2,rakke3;
  decl vecsigma,sigma;
  decl v_psi;
  decl v,c,cd;
  decl kappa,kappa_tilde;
  
  //empty matrix to results
  correction = zeros(1,1);

  //alternative procedures
  //1) estimate the unrestricted model
  i1_ana(0,4,0,0,1,rr,0,0,0,0);
  //2) estimate under the null
  //i1_ana(0,4,0,0,1,rr,0,0,1,restr);
  //3) use true values in simulation
  //i1_alpha        = a;
  //i1_beta         = b;
  //i1_gamma_mat[0] = g1;
  //i1_omega        = unit(5);
  ///////////////////////

  //we need the orthogonal complement
  i1_alpha_orto = nullspace(i1_alpha);
  
  //set constants
  ny = rr+(k-1)*p;
  s  = columns(restr)-(rows(zz2_star)-rows(zz2));
  nD = p_r;
  nd = p_u;

  //Setup matrix bigp and bigq
  bigp = unit(rr)+i1_beta'i1_alpha;
  bigq = i1_beta';
  for (i=1 ; i<=k-1 ; ++i)
    {
    bigp = bigp~i1_beta'i1_gamma_mat[i-1];
    }
  if(k>=2)
    {
    rakke2 = i1_alpha;
    for (i=1 ; i<=k-1 ; ++i)
      {
      rakke2 = rakke2~i1_gamma_mat[i-1];
      }
    bigp = bigp|rakke2;
    bigq = bigq|unit(p);
    }
  if(k>=3)
    {
    rakke3 = zeros((k-2)*p,rr)~unit((k-2)*p)~zeros((k-2)*p,p);
    bigp   = bigp|rakke3;
    bigq   = bigq|zeros((k-2)*p,p);
    }     

  //calculating parameters
  vecsigma    = invert(unit(ny^2)-(bigp**bigp))*vec(bigq*i1_omega*bigq');
  sigma       = shape(vecsigma,ny,ny);

  kappa       = choleski(invert(i1_alpha'invert(i1_omega)*i1_alpha));
  kappa_tilde = (unit(rr)|zeros((k-1)*p,rr))*kappa;
  v_psi       = kappa_tilde*kappa_tilde'invert(sigma);
  //alternative
  //v_psi       = trace( invert(unit(ny)-bigp)*bigq*i1_alpha*invert(i1_alpha'invert(i1_omega)*i1_alpha)*i1_alpha'bigq'invert(unit(ny)-bigp')*invert(sigma) );

  //calculating coefficients
  v    = trace(v_psi);
  c    = trace(bigp*invert(unit(ny)+bigp)*v_psi) + trace((bigp**((unit(ny)-bigp)*v_psi))*invert((unit(ny)**unit(ny))-(bigp**bigp)));

  //deterministic variables
  if(det==0)
    {
    //no deterministics
    cd = 0;
    }
  if(det==1)
    {
    //restricted constant
    cd = 0;
    }
  if(det==2)
    {
    //unrestricted constant
    cd = v;
    }
  if(det==3)
    {
    //restricted trend
    m  = 1;
    cd = trace((m**(unit(ny)-bigp)*v_psi)*invert((unit(nd)**unit(ny))-(m**bigp)));
    }   

  //calculating correction  
  correction[0][0] = 1+in*(0.5*(p+s-rr+1+2*nD)+nd+k*p)+(1/(n*rr))*((2*p+s-3*rr-1+2*nD)*v+2*(c+cd));
    
  //estimate under the null
  i1_ana(0,4,0,0,1,rr,0,0,1,restr);
  
  //calculate and corrected test
  btest = (n*cumulate(  (log((1-i1_egenv_r)./(1-i1_egenv_u[][0:columns(i1_egenv_r)-1])))'))[rr-1][0];
  btest = btest|(btest[0][0]/correction[0][0]);
  bpval = constant(1,2,1)-probchi(btest,fri*rr);

  //print results
  if(pr==1)
    {
    print("************************************************************************** \n \n");
    print("Test of hypothesis on Beta \n\n");
    print("Beta_star=B*Phi \n");
    print("B",restr);
    print("\nCase with ");
    if(det==0)
      {
      print("no deterministics");
      }
    else if(det==1)
      {
      print("restricted constant");
      }
    else if(det==2)
      {
      print("unrestricted constant");
      }
    else if(det==3)
      {
      print("restricted trend");
      }   

    print("\n\nTwo sets of eigenvalues");
    print("%r",{"","Unrestricted","Restricted"},
      (cumulate(ones(p,1))')|
      (i1_egenv_u[][0:p-1])|
      (i1_egenv_r[][0:p-1]));
  
    println("\n\n                       LR-test     p-value");
    println("Uncorrected            ",btest[0][0],"    ",bpval[0][0],"  in a Chi2(",fri*rr,")\n");
    println("Bartlett-correction    ",correction[0][0]);
    println("Bartlett-corrected     ",btest[1][0],"    ",bpval[1][0],"  in a Chi2(",fri*rr,")");
    }
  }