/*
**  Monte_Carlo_Replication_File.ox
**
**  Code for replicating results in
**  BOOTSTRAP TESTING OF HYPOTHESES ON CO-INTEGRATING RELATIONS IN VAR MODELS
**
**  This code replicates the Monte Carlo simulation results.
**  Written for OX 6.30 and requires the OxMPI/loop class, see www.doornik.com.
**
**  1) The relevant parameter configurations are constructed with the function setup_par(), see the file BootstrapPackage.ox.
**  2) The asymptotic LR test and the Bartlett correction is implemented with the function i1_lr_bartlett(), see the file i1_estimation.ox.
**  3) The proposed bootstrap test is implemented using the Bootstrap package, see the file BootstrapPackage.ox.
**     the test specifies EvaluateStatistic(data,isActualData) and GenerateSample(data,shock).
**  4) The alternative unrestricted bootstrap is specified by EvaluateStatistic_alternative() and GenerateSample().
**  5) The Monte Carlo simulation is implemented using the function FreeSimulation(), see the file BootstrapPackage.ox.
**     The simulation calls the function Simulation_Iteration(data) in each iteration.
**  6) The additional code in the main() function below collects the results for the size simulation or performs the pointwise size correction for power simulation.
**
**  Giuseppe Cavaliere, Heino Bohn Nielsen, and Anders Rahbek
**  University of Bologna and University of Copenhagen
**  University of Copenhagen
**  University of Copenhagen
*/


#include <i1_estimation.ox>
#include <BootstrapPackage.ox>


//
//global variables
//
decl dim,DGPmu;  
decl H,h;
decl H_original;
decl i_intercept;  


//
//Monte Carlo DGP for simulation
//
MonteCarloDGP(const par)
  {
  decl alpha,beta,pi,C,obs;

  //parameters
  alpha = par[0][0]|par[1][0]|zeros(dim-2,1);
  beta  = 1|par[2][0]|zeros(dim-2,1);
  pi    = alpha*beta'+unit(dim);

  //covariance
  C = unit(dim);

  //observations
  obs   = floor(exp(par[3][]));

  decl Y = constant(0,dim,obs);
  for(decl i=1;i<obs;++i)
    {
    //Y[][i] = DGPmu*<0;0;1;1> + pi*Y[][i-1] + C*rann(dim,1);
    Y[][i] = pi*Y[][i-1] + C*rann(dim,1);
    }

  return Y; 
  }


//
//evaluate statistic with BOOTSTRAP DGP ESTIMATED UNDER THE NULL
//
//  if isActualData==TRUE, also return residuals and estimates
//
EvaluateStatistic(const data,const isActualData)
  {
  decl stat,residual,residual_u,parameters;

  //estimation
  xx       = data;
  x1       = 0;
  if(i_intercept==0)
    {
    dx1      = 0;
    }
  else
    {
    dx1      = ones(1,columns(data));
    }

  //Supply information for the estimation   
  startper =    {1,1};
  freq     =        1;
  k        =        1;
  start    =      k+1; 
  end      =   columns(xx);
  
  //construct data matrices
  i1_datamat(0);

  if(isActualData==TRUE)
    {
    //also estimate under H
    i1_ana(0,4,0,3,1,1,0,0,0,0);
    //unrestricted residuals
    //residual_u = zeros(p,k)~conc(i1_residual,ones(1,n));

    i1_ana(0,4,0,3,1,1,0,0,1,H_original);
    test(0,i1_egenv_r,i1_egenv_u);

    //statistic
    stat = btest[0][0];

    //restricted residuals
    residual = zeros(p,k)~conc(i1_residual,ones(1,n));

    //parameters under null
    parameters = {i1_pi,i1_exo};

    //return
    return {stat , residual , parameters };
    }
  else
    {
    //only test
    i1_ana(0,4,0,3,1,1,0,0,0,0);
    i1_ana(0,4,0,3,1,1,0,0,1,H_original);
    test(0,i1_egenv_r,i1_egenv_u);

    //statistic
    stat = btest[0][0];
    
    //return
    return {stat , 0 , 0 };
    }       
  }

  
//
//evaluate statistic with BOOTSTRAP DGP ESTIMATED UNDER ALTERNATIVE
//
//  if isActualData==TRUE, also return residuals and estimates
//
EvaluateStatistic_alternative(const data,const isActualData)
  {
  decl stat,residual,parameters;

  //estimation
  xx       = data;
  x1       = 0;
  if(i_intercept==0)
    {
    dx1      = 0;
    }
  else
    {
    dx1      = ones(1,columns(data));
    }

  //Supply information for the estimation   
  startper =    {1,1};
  freq     =        1;
  k        =        1;
  start    =      k+1; 
  end      =   columns(xx);
  
  //construct data matrices
  i1_datamat(0);

  if(isActualData==TRUE)
    {
    //this is the original restriction:
    H = H_original;

    //unrestricted estimation
    i1_ana(0,4,0,3,1,1,0,0,0,0);

    //save parameters under ALTERNATIVE
    parameters = {i1_pi,i1_exo};
    decl beta_u =  i1_beta./i1_beta[0][0];

    //residuals under ALTERNATIVE
    residual = zeros(p,k)~conc(i1_residual,ones(1,n));
    
    //also estimate under H
    i1_ana(0,4,0,3,1,1,0,0,1,H);
    test(0,i1_egenv_r,i1_egenv_u);

    //statistic
    stat = btest[0][0];

    //this is the restriction to be testet in bootstrap replications: beta = beta_hat
    H = beta_u;

    //return
    return {stat , residual , parameters };
    }
  else
    {
    //only test
    i1_ana(0,4,0,3,1,1,0,0,0,0);
    i1_ana(0,4,0,3,1,1,0,0,1,H);
    test(0,i1_egenv_r,i1_egenv_u);

    //statistic
    stat = btest[0][0];
    
    //return
    return {stat , 0 , 0 };
    }       
  }


//
//generate a new bootstrap sample
//
//  may use function Bootstrap::GetParameters() to get saved parameters
//
GenerateSample(const data,const shock)
  {
  decl sample,pi,mu,dx;

  //parameters 
  pi     = Bootstrap::GetParameters()[0];
  if(i_intercept==1)
    {
    mu     = Bootstrap::GetParameters()[1];
    }

  //generate
  sample = data;
  for(decl i=1;i<columns(data);++i)
    {
    if(i_intercept==0)
      {
      dx          = pi*sample[][i-1] + shock[][i];
      }
    else
      {
      dx          = mu + pi*sample[][i-1] + shock[][i];
      }
    
    sample[][i] = dx + sample[][i-1];
    }

  //return  
  return sample;  
  }

  
//
//one Monte Carlo iteration
//
//  may use Bootstrap::Bootstrap_Test() to implement bootstrap
//
Simulation_Iteration(const data)
  {
  decl stat,statcor,statcor2,statcor3,collect;

  //estimation
  xx       = data;
  x1       = 0;
  if(i_intercept==0)
    {
    dx1      = 0;
    }
  else
    {
    dx1      = ones(1,columns(data));
    }

  //Supply information for the estimation   
  startper =    {1,1};
  freq     =        1;
  k        =        1;
  start    =        k+1; 
  end      =   columns(xx);
  
  //data matrices
  i1_datamat(0);


  //Asymptotic and Bartlett corrected test. Correction depends in deterministics
  if(i_intercept==0)
    {
    i1_lr_bartlett(0,1,0,H_original);
    }
  else
    {
    i1_lr_bartlett(0,1,2,H_original);
    }
  //collect all results 
  stat    = btest[0][0]|bpval[0][0]|(bpval[0][0]<=0.05)|(bpval[0][0]<=0.10)|btest[1][0]|bpval[1][0]|(bpval[1][0]<=0.05)|(bpval[1][0]<=0.10);
  collect = stat;


  //Bootstrap test with DGP estimated under the NULL
  decl myseed = ranseed(0);
  Bootstrap::Bootstrap_Test(0,"noMPI","iid",999,data,EvaluateStatistic,0,GenerateSample);
  stat    = Bootstrap::GetBootstrapResults();
  statcor = stat[1][1]|stat[3][0]|(stat[3][0]<=0.05)|(stat[3][0]<=0.10);
  collect = collect|statcor;


  //Bootstrap test with DGP estimated under the ALTERNATIVE
  ranseed(myseed);
  Bootstrap::Bootstrap_Test(0,"noMPI","iid",999,data,EvaluateStatistic_alternative,0,GenerateSample);
  stat    = Bootstrap::GetBootstrapResults();
  statcor = stat[1][1]|stat[3][0]|(stat[3][0]<=0.05)|(stat[3][0]<=0.10);
  collect = collect|statcor;


  //return
  return collect;
  }


  
///////////////////////////////////////////////////////////////////////  
//  
//main program block
//
main()
  {
  decl Y,help,datastar;
  decl par;
  decl t1,t2;

  //set deterministics  
  i_intercept = 0;   
  DGPmu       = 0;
  print("\nIntercept=",i_intercept);
  print("\nDGPmu=",DGPmu);
  
  //start timer
  t1=timer();
  format(2000);

  //set parameters
  dim = 4;
  par = <-0.5;
          0.0;
          0.0>|log(100.1);

  //one time series
  Y = MonteCarloDGP(par);

  //tau
  H = dropc(unit(dim),<1,2,3>);

  //Simulation_Iteration(Y);
  //EvaluateStatistic(Y,0);  
  //decl Boot = new Bootstrap();
  //Boot.Bootstrap_Test(1,"noMPI","iid",99,Y,EvaluateStatistic,{"LR"},GenerateSample);
  //decl stat = Bootstrap::GetBootstrapResults();
  
  H_original = H;
  print("\ntau",H);

  
  //size simulation
  par = Bootstrap::setup_par({-0.9|-0.05|0.05,
                               0.0| 0.9 |0.1,
                              <0.0>,
                              log(<100.1>)});

                              
  //power: difficult point                            
  //par = Bootstrap::setup_par({<-0.4>,
  //                            < 0.0>,
  //                              0.0| 0.10 |0.01,
  //                            log(<60.1>)})~
  //      Bootstrap::setup_par({<-0.4>,
  //                            < 0.0>,
  //                              0.15|2.00 |0.05,
  //                            log(<60.1>)});

  
  //power: easy point                             
  //par = Bootstrap::setup_par({<-0.8>,
  //                            < 0.8>,
  //                              0.000| 0.025 |0.005,
  //                            log(<60.1>)})~
  //      Bootstrap::setup_par({<-0.8>,
  //                            < 0.8>,
  //                              0.0| 0.30 |0.01,
  //                            log(<60.1>)})[][3:];

  
  //power, difficult point, T=100
  //par = Bootstrap::setup_par({<-0.4>,
  //                            < 0.0>,
  //                             0|0.25|0.01,
  //                            log(<100.1>)})~
  //      Bootstrap::setup_par({<-0.4>,
  //                            < 0.0>,
  //                             0.275|0.6|0.025,
  //                            log(<100.1>)});


  
  //size versus T
  //par = Bootstrap::setup_par({<-0.4>,
  //                            < 0.0>,
  //                             <0>,
  //                            log(40)|log(1000.1)|((log(1000.1)-log(40))/25)});
  //print( par'~floor(exp(par[3][]'))~log(floor(exp(par[3][]'))) );



  ////////////////////////
  //Monte carlo simulation
  //

  oxwarning(0);  
  decl Boot = new Bootstrap();
  Boot.FreeSimulation(FALSE,FALSE,10000,MonteCarloDGP,par,Simulation_Iteration,
  {"LR","p_LR","R5_LR","R10_LR",
   "Bart","p_Bart","R5_Bart","R10_Bart",
   "M_Boot","p_Boot","R5_Boot","R10_Boot",
   "M_ABoot","p_ABoot","R5_ABoot","R10_ABoot"});


  //
  //collect the results for the size simulation
  //
  if(Boot.GetSimulation_pr()==1)
    {
    decl cc = Boot.GetCollectedResults();
    decl collect   = constant(.NaN,columns(par),columns(cc[0]));
    for(decl i=0;i<columns(par);++i)
      {
      collect[i][]   = cc[i][1][];
      }
    print("\n\nCollected results",par'~collect);
    }

    
  //
  //pointwise size correct and collect results for power simulation
  //
  //decl cv5,cv10;
  //if(Boot.GetSimulation_pr()==1)
  //  {
  //  decl cc = Boot.GetCollectedResults();
  //  decl collect   = constant(.NaN,columns(par),columns(cc[0]));
  //  decl collect5  = constant(.NaN,columns(par),columns(cc[0]));
  //  decl collect10 = constant(.NaN,columns(par),columns(cc[0]));
  //  for(decl i=0;i<columns(par);++i)
  //    {
  // if(i==0)
  //      {
  //      cv5  = quantiler(Boot.GetFullResults()[i]',<.05>)';
  //      cv10 = quantiler(Boot.GetFullResults()[i]',<.10>)';
  //      }
  //    collect[i][]   = cc[i][1][];
  //    //print(Boot.GetFullResults()[i],cv5);
  //    //print(Boot.GetFullResults()[i].<=cv5);
  //    collect5[i][]  = meanc(Boot.GetFullResults()[i].<cv5);
  //    collect10[i][] = meanc(Boot.GetFullResults()[i].<cv10);
  //    }
  //  print("\n\nCollected results",par'~collect);
  //  print("\n\nSize corrected results",par'~collect5);
  //  print("\n\nSize corrected results",par'~collect10);
  //
  //  print("\n\nCollected-Collected results",par'~collect
  //        ~collect5[][1]~collect5[][5]~collect5[][9]~collect5[][13]~collect5[][17]
  //        ~collect10[][1]~collect10[][5]~collect10[][9]~collect10[][13]~collect10[][17]
  //        );
  //  }

  
  //time spent  
  t2=timer();
  print("\n\nElapsed time: ",timespan(t1,t2),"\n");
  }
                              